.so /mnt/dick/style3.nroff
.nr HA 1 Cycle the page headings.
.nr if 0 Do not include figures.
.pn 0
.ce
-*-
.sp 20
.ce 2
CAC Document No. 212
.sp
AN NROFF STYLE MANUAL
.sp 2
.rr
.dt
.bp 1
.ls 2
.ce
AN NROFF STYLE MACRO FILE
.pt "An Nroff Style Macro File"
.rr
.h1 "A STYLE FILE"
.pg
The nroff text processer on the Unix operating system is an extremely
powerful and general program for producing documents.
Unfortunately, the philosophy behind nroff was one of providing
a few powerful primitives which the user is expected to build into
his or her own document processing language.
While such diversity is laudable, the fact remains that the construction
of a personal document processing language is an arduous and expensive
process.
This document describes a file of nroff macros which may be used to
format documents in the style given in the American Psychological
Association Publication Manual.
The APA publications manual provides a consistent style, one which
is similar to that of many other scientific publication manuals.
In some cases, scientific societies have not issued any standard
publication manual and it is everyone for themselves.
In such cases, an author will generally succeed by submitting a manuscript
in a consistent and readable style.
Thus, this file will be of use to those who do not ordinarily use
or conform to the APA Style Manual.
.pg
The macro file described here provides an additional system of numbered
paragraphs which may have several levels of numbering, as well as optional paragraph
headings.
The file also provides an automatic table of contents for documents using
its system of section headings.
.h1 "NROFF CONVENTIONS"
.pg
Nroff has some syntactic conventions which cannot be circumvented.
All macros and function calls are required to be one or two letters
long.
As a result, it is not possible to assign macro names with any
intrinsic meaning.
Macros which were written fairly early in the development of the file
have names with some nmemonic significance.
However, most macros are named in arbitrary sequence.
For instance, all macros having to do with a table of contents
consist of two capital letters the first of which is "C".
Thus, the user will often be required to use macros in the document
which are arbitrarily named.
.h1 "UNIX AS A DOCUMENT LIBRARY"
.pg
This discussion will be primarily of how the style file is used rather
than of how it operates internally.
In some cases the internals of the file are somewhat more elaborate
than is necessary, primarily to allow the later inclusion of new features.
Extensions to the file will be discussed later.
.h4     "Maintaining document files."
It is important to develop a file naming discipline early in the use
of Unix.
Without such a discipline, the user becomes hopelessly lost in a welter
of files with
.ul
ad hoc
names whose meaning has been forgotten.
The hierarchial structure of the Unix file system makes the development
of file naming conventions somewhat easier than is the case on machines
with only a single level of files.
The file naming conventions discussed here were developed over a period
of several years and have proved serviceable over a seemingly endless
series of documents.
The reader is encouraged to be innovative in developing a file naming
discipline, but could do worse than to start with the one here.
.pg
Begin by creating a directory with a short descriptive name for the document.
For example, an article on turnover rate in working class housing might
be put in a directory called "turnov".
The first version of the article would be called "turnov/a1.nroff".
The fairly long name of the directory serves as an umbrella covering all
files that belong to this paper.
The file name "a1.nroff" shows that this is version 1 of part "a" of the paper,
and that it is an nroff manuscript.
A revision of this file would be kept in turnov/a2.nroff, while a second
chapter or section of the paper would begin in turnov/b1.nroff.
Suffixes to files should show the program for which the file is destined,
not the program from which it came.
It is of no concern whether a printer-image file was produced by hand
or by a document processer; what is important is that the file is destined
for the printer rather than some where else.
.h4     "Keeping successive versions."
A Unix author should make a practice of maintaining successive versions of
a document in different and appropriately named files.
It is often possible to make a propagating error using a document editor,
and the use of successive file versions will allow recovery from such
errors.
In general, whenever a non-trivial editing session is begun, the latest
file should be copied to a file one version higher and the editing done
on the higher numbered version.
The Unix write-up for the program "diff" shows a procedure for comparing
successive versions and for keeping highly compressed copies of earlier
versions of a document.
.h3  "Using the Style Macro File"
.pg
All documents which use this macro file should begin with the line
     .so /mnt/dick/style/styl4.n
.br
which will automatically insert the style file into the beginning of
the document.
(It is good practice to use the full name of the file, so that it is
found regardless of which directory the user is currently using.)
This insertion will automatically set up a page of sixty spaces offset twelve
spaces from the left-hand side of an eleven inch page.
The document will automatically be page numbered and has room for
page headings.
If a double spaced document is desired, follow the ".so" line with a
line containing ".ls 2".
.h4     "Page titles."
Page titles may be put in with a special macro call.
To put page titles on page 2 and following pages, insert a line .pt "[first line]"
"[second line]" (the brackets are not part of the strings, but are
only to indicate the content of the strings.)
Both the first and second strings in the .pt macro are optional.
The first string will appear on the same line as the page number, and
the second string immediately below it.
The .pt macro was used to create the page titles for this document.
The macro file also contains a switch which can be used to set the page
titles into an alternating mode, which will facilitate printing on two
sides of the paper.
To alternate the titles, insert a line saying .nr HA 1 which will set
the switch register HA to 1.
If page numbers are to appear only on the right-hand side of the page,
then this line
may be omitted, or may be reversed with the command .nr HA 0.
.h4     "Dates."
The macro .dt will automatically provide a centered date of the form
"11 June 1976".
The current date in this form can also be used within the document by
using the string "\\*(dc".
If this date string is to be used in a header such as in the .pt macro
above, then it should be entered as "\\\\*(dc", which will supress its
interpretation until the macro is executed.
.h4     "Author line."
This file contains a macro ".rr", which contains the author's name and
affiliation.
The reader should delete this macro, but use its style to provide his or
her own author line.
.h4     "Paragraph."
The command ".pg" will begin a standard indented paragraph.
If fewer than two lines remain on the present page, nroff will begin
a new page.
.h3  "APA Section Headings"
.pg
The following set of heading conforms to the requirements of the American
Psychological Style Manual.
It is suggested that this style be followed for all journals which do not
explicitly state their style requirements.
The headings are simple and consistent and are easily translated into
type set.
In addition, use of these headings will automatically result in the
generation of a table of contents.
.h4     "First order headings."
A first order heading is produced by the commands
\ \ \ \ \ .h1 "[TITLE]"
.br
where the title in the brackets is the title
of the section.
First order headings are defined as centered upper-case text used to
introduce a major section of a paper or chapter.
The first heading of this paper was generated using the ".h1" command.
.h4     "Second order headings."
A second order heading is put in with the command
\ \ \ \ \ .h2 "[Title of Section]"
.br
a second order heading is capitalized, underlined, and centered.
Second order headings are set in italic type.
No second order headings occur in this document because this level
of heading is not used until the capacity of first, third, and fourth
order headings have been exhaused.
.h4     "Third order headings."
Third order headings are produced with the command
\ \ \ \ \ .h3 "[Title of Section]
.br
third order headings are capitalized italics beginning at the left margin
of the page.
Third order headings, like all underlined text in a typescript, are
set in italic type.
The section on headings is a third order heading produced with the
".h3" command.
.h4     "Fourth order headings."
Fourth order headings are produced by the command
\ \ \ \ \ .h4 "[Title of section.]"
.br
where the title is in sentence form with the first letter capitalized and
followed by a period.
The header to this paragraph was produced with a ".h4" command.
The fourth level headings need not be followed by a ".pg" command since
it is itself a paragraph.
However, all higher level headings must be explicitly followed by a paragraph
command.
.h3  "Choosing a Level of Header"
.pg
The APA Publications Manual suggests that, if only one level of header is
required, a first level be used.
This should be followed by the use of fourth order headings and then
third order headings.
Only in exceptionally long or intricate documents should a second order
heading be used.
.h3  "Numbered Paragraphs"
.pg
An alternate system of headings, much beloved by engineers and technical
people, is the use of paragraphs with numbers
of the form N.n.n.
These numbers can be produced automatically by the use of a set of
commands called ".p1", ".p2", ".p3", and ".p4".
Each of these commands will take an optional section title and paragraph
number.
The use of the command
\ \ \ \ \ .p1 "[title of section]"
.br
will result in a paragraph beginning "1. title of section".
The second time ".p1" is used, a paragraph beginning with "2. [title]"
will be generated.
.pg
The command ".p2" with an optional title, will generate a paragraph
beginning "N.1", where N is the number of the paragraph generated by
".p1".
.pg
The commands ".p3" and ".p4" can be used to generate paragraphs beginning
with three and four level numbers.
Each time a higher level paragraph is generated, all lower level paragraphs
are renumbered to begin at one.
The result is a hierarchial set of paragraphs whose numbers are automatically
maintained and incremented.
It is possible to begin the numbering sequence at any point by putting
an optional paragraph number after the title.
Thus, the command
\ \ \ \ \ .p3 "Methods of procedure" 2 4
.br
could be used to generate the title "2.4.1 Methods of procedure" as the
first numbered paragraph of a document.
It is not necessary to initialize the third level number to 1 since this
is done automatically.
The next use of a ".p3" command in the document would produce paragraph
2.4.2 while the next use of a ".p2" command would produce a paragraph
headed 2.5.1.
Remember to include a blank title when initializing a set of numbered
paragraphs which are not to carry explicit titles.
Otherwise the first of the initialization parameters would be used as the
paragraph title.
.h3  "Figure Insertion"
.pg
The commands ".FB" will generate a centered reference saying INCLUDE
TABLE N ABOUT HERE, where N is a figure number which is automatically
incremented each time this reference is used.
If desired, the figure number may be reset by adding it explicitly
as an optional parameter to the command.
Thus, it is possible to begin a file with Figure 7 by using the command
".FB 7".
The next use of the ".FB" command without a parameter will result in a
reference to Figure 8.
A later release of this macro file will have a switch called FA which
will automatically either insert figure references or the figures
themselves.
However, these procedures are not yet reliable enough to distribute.
.pg
The command ".TB" can be used in the same fashion to generate a table
reference like the figure reference.
.h3  "Footnotes"
.pg
Footnotes can be included in the body of the text preceeded by the ".fn"
command.
The end of the footnote text must be marked with a ".ef" command.
When the document is printed, the text contains a footnote number at
the point where the ".fn" command appeared, while the footnote or footnotes
are placed at the bottom of the page.
The ".fn" command takes an optional parameter, which is the word to which
the footnote reference will be tied.
Use of this feature makes it possible to have the footnote reference
centered when it is attached to a title and prevents the footnote reference
from being split off from the sentence to which it applies and placed on a
new line.
The sequence
\ \ \ \ \ .ce
\ \ \ \ \ .fn "How We Did It"
.br
\ \ \ \ \ Supported by the Bureau of Winds and Tides."
\ \ \ \ \ .ef
.br
would produce the centered title, "How We Did It[1]"
with an attached footnote reference, and
would put the statement of support at the bottom of the page.
.h3  "Table of Contents"
.pg
A table of contents will automatically be generated if the document is
ended with the command ".CA".
The table of contents will consist of all headers and numbered paragraphs
and their page numbers.
The table of contents will be produced on Roman Numbered pages with the
current page title.
The title will alternate if the alternation switch has been set.
.h1 "EXTENSIONS OF THE STYLE FILE"
.pg
A number of extensions of the file described here are in progress.
The major extension is to produce figures and tables in "floating keeps"
which will be inserted in the document when they will fit onto a page.
In addition, separate tables of contents for figures and tables should
be provided.
Finally, facilities will be added to produce lists of footnotes at the
end of a chapter or document as well as at the bottom of a page.
.bp
.h1 "TEXT OF THE STYLE MACRO FILE"
.pg
The current text of the style file can be seen by printing
"/mnt/dick/style/styl4.n", which is also the file to be sent to nroff
ahead of your text.
.CA
.nr CG 0
.ls 2
.bp 3
.fi
.ce
AN NROFF STYLE MACRO FILE
.rr
.ce
ABSTRACT
.sp 2
.pg
A set of style macros for the nroff text processer running on the
PDP 11 under Unix is described.
The style macros provide headings which conform to the American
Psychological Association Style Manual.
The macros provide facilities including automatic page numbering
and titling, a consistent system of section headings, automatic
figure reference and insertion, automatic table reference and
insertion.
The macros also provide for automatic generation of a table of contents.
The file contains numerous utility functions.
